#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string>
#include <stdio.h>
#include <ctype.h>
#include <iostream>

#include "global.h"
#include "symtab.h"

using namespace std;

#define SYMBOL_TABLE_LENGTH 1000
#define PARAM_TABLE_LENGTH 1000


// An entry in the symbol table
typedef struct symEntry_s
{
	string Name;
	int Type;
	string InitialValue;
	int NumParams;
} symEntry_t;


symEntry_t SymbolTable[SYMBOL_TABLE_LENGTH];
int NumSyms = 0;


symEntry_t ParamTable[SYMBOL_TABLE_LENGTH];
int NumParams = 0;



void DumpTable(void)
{
	cout << "Symbols: " << endl;
	for (int i = 0; i < NumSyms; i++)
	{
		cout << SymbolTable[i].Name << " ";
		switch (SymbolTable[i].Type)
		{
			case TYPE_UNDEF:
				Abort ("Symbol table is corrupt");
				break;
			case TYPE_BYTE:
				cout << "byte";
				break;
			case TYPE_WORD:
				cout << "word";
				break;
			case TYPE_LONG:
				cout << "long";
				break;
			default:
				Abort ("Symbol table is corrupt");
				break;
		}
		cout << endl;
	}
}


// Report the type of a symbol table entry
int SymbolType (string Name)
{
	for (int i = 0; i < NumSyms; i++)
	{
		if (SymbolTable[i].Name == Name)
		{
			return (SymbolTable[i].Type);
		}
	}
	return (TYPE_UNDEF);
}




// Report if a variable is in the table
bool InTable(string Name)
{
	for (int i = 0; i < NumSyms; i++)
	{
		if (SymbolTable[i].Name == Name)
		{
			return(true);
		}
	}
	return (false);
}


// Check for a duplicate variable name
// Report an error if identifier is already in table
void CheckDup(string Name)
{
	if (InTable(Name))
	{
		Abort(string("Duplicate Name") + Name);
	}
}


// Add a new entry to the symbol table
void AddEntry(string Name, int Type)
{
	CheckDup(Name);
	if (NumSyms == SYMBOL_TABLE_LENGTH)
	{
		Abort("Symbol table full");
	}
	SymbolTable[NumSyms].Name = Name;
	SymbolTable[NumSyms].Type = Type;
	NumSyms++;
}


// Check if a symbol is in the table, report an error if not
void CheckTable(string Symbol)
{
	if (!InTable(Symbol))
	{
		Undefined(Symbol);
	}
}


// Find the parameter number
int ParamNumber(string N)
{
	for (int i = 0; i < NumParams; i++)
	{
		if (ParamTable[i].Name == N)
		{
			return (i + 1);
		}
	}
	Abort ("Parameter missing from table");
	return (-1);
}


// Find the type of a parameter
int ParamType (string Name)
{
	for (int i = 0; i < NumParams; i++)
	{
		if (ParamTable[i].Name == Name)
		{
			return (ParamTable[i].Type);
		}
	}
	return (TYPE_UNDEF);
}


// See if an identifier is a parameter
bool IsParam(string Name)
{
	for (int i = 0; i < NumParams; i++)
	{
		if (Name == ParamTable[i].Name)
		{
			return(true);
		}
	}
	return (false);
}


// Add a new parameter to the table
void AddParam (string Name, int Type, string InitialValue)
{
	if (IsParam(Name))
	{
		Duplicate(Name);
	}
	if (NumParams == PARAM_TABLE_LENGTH)
	{
		Abort ("Param table full");
	}
	ParamTable[NumParams].Name = Name;
	ParamTable[NumParams].Type = Type;
	ParamTable[NumParams].InitialValue = InitialValue;
	NumParams++;
}


// Initialize parameter table to null
void ClearParams()
{
	for (int i = 0; i < NumParams; i++)
	{
		ParamTable[i].Name = "";
		ParamTable[i].Type = TYPE_UNDEF;
	}
	NumParams = 0;	
}


// Return type of variable from param table or sym table
int TypeOf (string Name)
{
	int Type;

	Type = ParamType(Name);
	if (Type == TYPE_UNDEF)
	{
		return SymbolType(Name);
	}
	return Type;
}


// Set number of params of a proc
void SetNumParams (string Name, int NumParams)
{
	for (int i = 0; i < NumSyms; i++)
	{
		if (SymbolTable[i].Name == Name)
		{
			SymbolTable[i].NumParams = NumParams;
			return;
		}
	}
	Abort ("Internal compiler error");
}

// Check number of params of a proc
void CheckNumParams (string Name, int NumParams)
{
	for (int i = 0; i < NumSyms; i++)
	{
		if (SymbolTable[i].Name == Name)
		{
			if (SymbolTable[i].NumParams != NumParams)
			{
				Abort ("Incorrect number of parameters for call");
			}
			return;
		}
	}
	Abort ("Internal compiler error");
}



