#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string>
#include <stdio.h>
#include <ctype.h>
#include <iostream>

#include "global.h"
#include "output.h"
#include "symtab.h"

// Do we want to enforce use of semicolons ?
//#define MANDATORY_SEMICOLONS 1

#define SCOPE_GLOBAL 0
#define SCOPE_LOCAL  1

using namespace std;


// Type declarations
typedef string Symbol;
typedef Symbol* TabPtr;

// Input file
int fd;

// Tokens are ints but operator ASCII values are used for the operators
// so use non-ascii values for ident and keyword tokens
#define TOK_NUM		0x0100
#define TOK_ID		0x0101
#define TOK_IF		0x0102
#define TOK_ELSE	0x0103
#define TOK_ENDIF	0x0104
#define TOK_END		0x0104	// N.B. Same as endwhile and endif
#define TOK_ENDWHILE	0x0104	// N.B. Same as endif and end
#define TOK_WHILE	0x0105
#define TOK_READ	0x0106
#define TOK_WRITE	0x0107
#define TOK_BYTE	0x0108
#define TOK_WORD	0x0109
#define TOK_LONG	0x010a
#define TOK_UBYTE	0x010b
#define TOK_UWORD	0x010c
#define TOK_ULONG	0x010d
#define TOK_PROC	0x010e
#define TOK_FUNC	0x010f


#define NKW1 	17
#define NKW 	16
int KWCode[NKW1] = { TOK_ID, TOK_IF,       TOK_ELSE, TOK_ENDIF, TOK_WHILE,
                             TOK_ENDWHILE, TOK_READ, TOK_WRITE, TOK_BYTE,  
                             TOK_WORD,     TOK_LONG, TOK_UBYTE, TOK_UWORD,
			     TOK_ULONG,    TOK_PROC, TOK_FUNC,	TOK_END};

// Definitions of keyword and Token Types
#undef SUOMI
#ifdef SUOMI
Symbol KWList[NKW] = {	"jos",		"else",		"loppujos", 	"kuntaas",
			"loppukuntaas",	"read",		"write",	"sbyte",
			"sword",	"slong",	"ubyte",	"uword",
			"ulong",	"kulku",	"func",		"loppu"};
			
#define BLOCK_START	"alku"
#define BLOCK_END	"loppu"
#define ENDIF		"loppujos"
#define ENDWHILE	"loppukuntaas"
#else
Symbol KWList[NKW] = {	"if",		"else", 	"endif", 	"while",
			"endwhile",	"read", 	"write", 	"sbyte",
			"sword",	"slong",	"ubyte",	"uword",
			"ulong",	"proc",		"func",		"end"};
			
#define BLOCK_START	"begin"
#define BLOCK_END	"end"
#define ENDIF		"endif"
#define ENDWHILE	"endwhile"
#endif



char Look;		// Lookahead Character

int Token;		// Output from lexical scanner

string Value;


int Base = 0;

int LineNumber = 0;

void log (string s)
{
	cerr << s << ": " << "Token = " << Token << " Value = " << Value << " Look = " << Look << endl;
}


// Report an Error
void Error(string s)
{
	cerr << "Error: " << s << "." << endl;
}



// Report Error and Halt
void Abort(string s)
{
	fprintf (stderr, "Line %d: ", LineNumber);
	Error(s);
	exit(1);
};


// Report Warning
void Warning(string s)
{
	fprintf (stderr, "Line %d: ", LineNumber);
	cerr << "Warning: " << s << "." << endl;
};



// Read New Character From Input Stream, result in global "look"
void GetChar()
{
	int len;
	len = read (fd, &Look, 1);
	if (len == 1)
	{
		if (Look == '\n')
		{
			LineNumber++;
		}
	}
	else
	{
		perror("");
		Abort ("GetChar: cannot read");
	}
}


// Skip a comment field
void SkipComment()
{
	while (Look != '}')
	{
		GetChar();
		if (Look == '{')
		{
			SkipComment();
		}
	}
	GetChar();
}


// Report what was Expected
void Expected(string s)
{
	Abort("Expected: " + s + "");
}


// Report an undefine identifier
void Undefined (string s)
{
	Abort ("Undefined identifier: " + s);
}


// Report a duplicate identifier
void Duplicate (string s)
{
	Abort ("Duplicate identifier: " + s);
}


// Check that the current token is an identifier
void CheckIdent()
{
	if (Token != TOK_ID)
	{
		Expected ("Identifier");
	}
}


// Recognize an Alpha Character
bool IsAlpha(char c)
{
	return (isalpha(c));
}
                              


// Recognize a Decimal Digit
bool IsDigit(char c)
{
	return (isdigit(c));
}


// Recognize an Alphanumeric
bool IsAlNum(char c)
{
	return (IsAlpha(c) | IsDigit(c));
}


// Recognize an add op ("+" or "-")
bool IsAddOp(char c)
{
	switch (c)
	{
		case '+':
		case '-':
			return(true);
			break;
		default:
			return(false);
	}
}


// Recognize a mul op ("*" or "/")
bool IsMulOp(char c)
{
	switch (c)
	{
		case '*':
		case '/':
			return(true);
			break;
		default:
			return(false);
	}
}


// Recognize a boolean orop ("|" or "~")
bool IsOrOp(char c)
{
	switch (c)
	{
		case '|':
		case '~':
			return(true);
			break;
		default:
			return(false);
	}
}


// Recognize a Relop
bool IsRelOp(char c)
{
	switch (c)
	{
		case '=':
		case '<':
		case '>':
			return (true);
			break;
		default:
			return (false);
			break;
	}
}


// Recognize White Space
bool IsWhite(char c)
{
	switch (c)
	{
		case '\n':
		case '\r':
		case ' ':
		case '\t':
		case '{':
			return (true);
		break;
		default:
			return (false);
		break;
	}
}


// Recognise a legal variable type
bool IsVarType(int Type)
{
	switch (Type)
	{
		case TYPE_BYTE:
		case TYPE_WORD:
		case TYPE_LONG:
		case TYPE_UBYTE:
		case TYPE_UWORD:
		case TYPE_ULONG:
			return (true);
			break;
		default:
			break;
	}
	return (false);
}

// Get a variable from the symbol table
int VarType (string Name)
{
	int Type = TypeOf(Name);
	if (!IsVarType(Type))
	{
		Abort ("Identifier " + Name + " is not a variable");
	}
	return (Type);
}




// Skip Over Leading White Space
void SkipWhite()
{
	while (IsWhite(Look))
	{
		if (Look == '{')
		{
			SkipComment();
		}
		GetChar();
	}
}


// Get an entifier
void GetName()
{
	SkipWhite();
	if (!IsAlpha(Look))
	{
		Expected("Identifier");
	}
	Token = TOK_ID;
	Value = "";
	do	
	{
		Value += Look;// (char)toupper(Look);
   		GetChar();
	}
	while (IsAlNum(Look) || Look == '_');
}


// Get a Number,
void  GetNum()
{
	SkipWhite();
	if (!IsDigit(Look))
	{
		Expected("Number");
	}
	Token = TOK_NUM;
	Value = "";
	do
	{
		Value = Value + Look;
		GetChar();
	}
	while (IsDigit(Look));
}


// Get a operator
void GetOp()
{
	SkipWhite();
	Token = Look;
	Value = Look;
	GetChar();
}


// Get the next input token
void Next()
{
	SkipWhite();
	if (IsAlpha(Look))
	{
		GetName();
	}
	else if (IsDigit(Look))
	{
		GetNum();
	}
	else
	{
		GetOp();
	}
}


// Table lookup
// If the input string is in the table return it's index
// If not return -1
int LookUp(TabPtr T, string s, int n)
{
	int ret = -1;
	for (int i = 0; i < n; i++)
	{
		if (s == T[i])
		{
			ret = i;
			break;
		}
	}
	return (ret);
}


// Scan the current identifier for keywords
void Scan()
{
	if (Token == TOK_ID)
	{
		int k = LookUp(KWList, Value, NKW);
		if (k == -1)
		{
			Token = TOK_ID;	
		}
		else
		{
			Token = KWCode[k + 1];
		}
	}
}


// Match a specific string
void MatchString (string x)
{
	if (Value != x)
	{
		Expected (x);
	}
	Next();
}


// Match a semicolon
void MatchSemi()
{
#ifdef MANDATORY_SEMICOLONS
	MatchString(";");
#else
	if (Token == ';')
	{
		Next();
	}
#endif
}


// Load a variable into the primary register
int Load(string Name)
{
	int Typ;

	Typ = VarType(Name);
	LoadVar(Name, Typ);
	return (Typ);
}


// Return the allocation size of a given type
int TypeSize(int Type)
{
	switch (Type)
	{
		case TYPE_BYTE:
		case TYPE_UBYTE:
			return (1);
			break;
		case TYPE_WORD:
		case TYPE_UWORD:
			return (2);
			break;
		case TYPE_LONG:
		case TYPE_ULONG:
			return (4);
			break;
		default:
			Abort ("TypeSize: Unknown type");
			break;
	}
	return (-1);
}


// Check if conversion between these types will loose information
void CheckTypeConvertion(int SrcType, int DstType)
{
	if (TypeSize(SrcType) > TypeSize(DstType))
	{
		Warning ("Type mismatch");
	}
}


// Store a variable from the primary register
void Store (string Name, int PrimType)
{
	int DstType;

	DstType = VarType(Name);
	//Convert(PrimType, DstType);
	CheckTypeConvertion(PrimType, DstType);
	StoreVar(Name, DstType);
}


int LoadNum (string Value)
{
	unsigned long long Num;
	int Type;

	Num = atoll(Value.c_str());

	if (Num <= 127)
	{
		Type = TYPE_BYTE;
	}
	else if (Num <= 32767)
	{
		Type = TYPE_WORD;
	}
	else
	{
		Type = TYPE_LONG;
	}
	LoadConst (Value, Type);
	return (Type);
}


int LoadParam (string Value)
{
	int Type;
	int ParNum;

	Type = ParamType(Value);
	ParNum = ParamNumber(Value);
	LoadParameter (ParNum, Type);
	return (Type);
}


int BoolExpression();
void CallProc (string Name);

// Parse and translate a math factor
int Factor()
{
	int Type;
	
	if (Token == '(')
	{
		Next();
		Type = BoolExpression();
		MatchString(")");
	}
	else
	{
		if (Token == TOK_ID)
		{
			if (Look == '(') // FIXME IsFunction(Value))
			{
				log ("Factor_1:");
				CallProc(Value); 	// FIXME this should be CallFunc (or Call)
				log ("Factor_2:");
				Type = TYPE_LONG;	// FIXME type should be from function type
			}
			else if (IsParam(Value))
			{
				Type = LoadParam (Value);
				Next();
			}
			else
			{
				Type = Load(Value);
				Next();
			}
		}	
		else if (Token == TOK_NUM)
		{
			Type = LoadNum(Value);
			Next();
		}
		else
		{
			Expected ("Math factor");
		}
	}
	return (Type);
}


int Multiply(int Type)
{
	int FactorType;
	Next();
	FactorType = Factor();
	return PopMul(Type, FactorType);
}


int Divide(int Type)
{
	int FactorType;
	Next();
	FactorType = Factor();
	return PopDiv(Type, FactorType);
}


int Term()
{
	int Type;
	
	Type = Factor();
	while (IsMulOp(Token))
	{
		Push(Type);
		switch (Token)
		{
			case '*':
				Type = Multiply(Type);
				break;
			case '/':
				Type = Divide(Type);
				break;
		}
	}
	return (Type);
}


// Recognize and Translate an Add
int Add(int Type)
{
	int TermType;

	Next();
	TermType = Term();
	return  PopAdd(Type, TermType);
}


// Recognize and Translate a Subtract
int Subtract(int Type)
{
	int TermType;
	
	Next();
	TermType = Term();
	return PopSub(Type, TermType);
}


// Parse and Translate an Expression
int Expression(int Type)
{
	if (IsAddOp(Token))
	{
		Type = Clear();
	}
	else
	{
		Type = Term();
	}
	while (IsAddOp(Token))
	{
		Push(Type);
		switch (Token)
		{
			case '+':
				Type = Add(Type);
				break;
			case '-':
				Type = Subtract(Type);
				break;
		}
	}
	return (Type);
}


// Get another expression and compare
int CompareExpression(int Type)
{
	int ExpType = Expression(Type);
	return PopCompare(Type, ExpType);
}


// Get the next expression and compare
int NextExpression(int Type)
{
	Next();
	return CompareExpression(Type);
}


// Recognize and Translate a Relational "Equal"
int Equal(int Type)
{
	Type = NextExpression(Type);
	SetEqual(Type);
	return (Type);
}


// Recognize and translate a relational "Less than or equal"
int LessOrEqual(int Type)
{
	Type = NextExpression(Type);
	SetLessOrEqual(Type);
	return (Type);
}



// Recognize and Translate a Relational "Not Equals"
int NotEqual(int Type)
{
	Type = NextExpression(Type);
	SetNotEqual(Type);
	return (Type);
}


// Recognize and Translate a Relational "Less Than"
int Less(int Type)
{
	Next();
	switch (Token)
	{
		case '=':
			Type = LessOrEqual(Type);
			break;
		case '>':
			Type = NotEqual(Type);
			break;
		default:
			Type = CompareExpression(Type);
			SetLess(Type);
			break;
	}
	return (Type);
}


// Recognize and Translate a Relational "Greater Than"
int Greater(int Type)
{
	Next();
	if (Token == '=')
	{
		Type = NextExpression(Type);
		SetGreaterOrEqual(Type);
	}
	else
	{
		Type = CompareExpression(Type);
		SetGreater(Type);
	}
	return (Type);
}


// Parse and Translate a Relation
int Relation(int Type)
{
	Type = Expression(Type);
	if (IsRelOp(Token))
	{
		Push(Type);
	        switch (Token)
		{
			case '=':
				Type = Equal(Type);
				break;
			case '<':
				Type = Less(Type);
				break;
			case '>':
				Type = Greater(Type);
				break;
			default:
				Abort ("Relation: unknown operator");
				break;
		}
	}
	return (Type);
}


// Parse and Translate a Boolean Factor with leading NOT
int NotFactor(int Type)
{
	if (Token == '!')
	{
		Next();
		Type = Relation(Type);
		NotIt(Type);
	}
	else
	{
		Type = Relation(Type);
	}
	return (Type);
}


// Parse and Translate a Boolean Term
int BoolTerm(int Type)
{	
	Type = NotFactor(Type);
	while (Token == '&')
	{
		Push(Type);
		Next();
		Type = PopAnd(Type, NotFactor(Type));
	}
	return (Type);
}


// Recognize and Translate a Boolean OR
int BoolOr(int Type)
{
	int BoolTermType;

	Next();
	BoolTermType = BoolTerm(Type);
	return PopOr(Type, BoolTermType);
}


// Recognize and Translate a Boolean XOR
int BoolXor(int Type)
{
	int BoolTermType;

	Next();
	BoolTermType = BoolTerm(Type);
	return PopXor(Type, BoolTermType);
}


// Parse and Translate a Boolean Expression
int BoolExpression()
{
	int Type;
	
	Type = BoolTerm(Type);
	while (IsOrOp(Token))
	{
		Push(Type);
		switch (Token)
		{
			case '|':
				Type = BoolOr(Type);
				break;
			case '~':
				Type = BoolXor(Type);
				break;
		}
	}
	return (Type);
}


// Parse and Translate an Assignment Statement
void Assignment(string Name)
{
	int Type;

	// CheckTable(Value); FIXME check is in AssignOrProc
	Next();
	log ("Assignment:");
	MatchString("=");
	Type = BoolExpression();
	if (IsParam(Name))
	{
		StoreParam(ParamNumber(Name));
	}
	else
	{
		Store(Name, Type);
	}
}


// Process an actual parameter
void Param()
{
	int Type;
	
	Type = BoolExpression();
	Push (Type);
}


// Process the parameter list of a call
int ParamList(string Name)
{
	int bytesPushed = 0;
	int NumParams = 0;
	
	MatchString("(");	
	if (Token != ')')
	{
		Param();
		NumParams++;
		bytesPushed += 4;
		while (Token == ',')
		{
			MatchString(",");
			Param();
			NumParams++;
			bytesPushed += 4;
		}
	}
	MatchString(")");
	CheckNumParams (Name, NumParams);
	return (bytesPushed);
}


// Process a procedure call
void CallProc(string Name)
{
	int bytesPushed;
	Next();
	bytesPushed = ParamList(Name);
	Call(Name);
	CleanStack(bytesPushed);
}


// Decide if a statement is an assignment or procedure call
void AssignOrProc()
{
	switch (TypeOf(Value))
	{
		case TYPE_UNDEF:
			Undefined(Value);
			break;
		case TYPE_BYTE:
		case TYPE_WORD:
		case TYPE_LONG:
		case TYPE_UBYTE:
		case TYPE_UWORD:
		case TYPE_ULONG:
			Assignment(Value);
			break;
		case TYPE_PROC:
			CallProc(Value);
			log ("AssignOrProc");
			break;
		default:
			Abort ("Identifer " + Value + "Cannot be used here");
			break;
	}
}


void DoBlock();

// Recognize and Translate an IF Construct
void DoIf()
{
	string* L1;
	string* L2;

	Next();
	BoolExpression();
	L1 = NewLabel();
	L2 = L1;
	BranchFalse(*L1);
	DoBlock();
	if (Token == TOK_ELSE)
	{
		Next();
		L2 = NewLabel();
		Branch(*L2);
		PostLabel(*L1);
		DoBlock();
	}
	PostLabel(*L2);
	MatchString(ENDIF);
}


// Parse and translate a WHILE statement
void DoWhile()
{
	string* L1;
	string* L2;
	
	Next();
	L1 = NewLabel();
        L2 = NewLabel();
	PostLabel(*L1);
	BoolExpression();
	BranchFalse(*L2);
	DoBlock();
	MatchString(ENDWHILE);
	Branch(*L1);
	PostLabel(*L2);
}


void BeginBlock();


// Process a formal parameter -- dummy just skips it
void FormalParam(int Type)
{
	string Name;

	Next();
	CheckIdent();
	AddParam(Value, Type, "");
}



// Process the formal parameter list of a procedure
void FormalList()
{
	MatchString("(");
	Scan();
	while ((Token == TOK_BYTE) || (Token == TOK_WORD) || (Token == TOK_LONG)
		|| (Token == TOK_UBYTE) || (Token == TOK_UWORD) || (Token == TOK_ULONG))
	{	
		switch (Token)
		{
			case TOK_BYTE:
				FormalParam(TYPE_BYTE);
				break;
			case TOK_UBYTE:
				FormalParam(TYPE_UBYTE);
				break;
			case TOK_WORD:
				FormalParam(TYPE_WORD);
				break;
			case TOK_UWORD:
				FormalParam(TYPE_UWORD);
				break;
			case TOK_LONG:
				FormalParam(TYPE_LONG);
				break;
			case TOK_ULONG:
				FormalParam(TYPE_ULONG);
				break;
			default:
				Expected ("type");
				break;
		}
		Next();
		if (Token == ',')
		{
			MatchString(",");
		}
		Scan();
	}
	MatchString(")");
	Base = NumParams;
	NumParams = NumParams + 2;
}

int Decls(int Scope);

// Parse and translate a procedure declaration
void DoProc()
{
	string Name;
	int k;

	Next();
	CheckIdent();	
	CheckDup(Value);
	AddEntry(Value, TYPE_PROC);
	Name = Value;
	Next();
	FormalList();
	SetNumParams(Name, Base);
	k = Decls(SCOPE_LOCAL);
	ProcProlog(Name, k);
	BeginBlock();
	ProcEpilog(Name);
	ClearParams();
}


// Parse and translate a function declaration FIXME This is same a proc
void DoFunc()
{
	string Name;
	int k;

	Next();
	CheckIdent();	
	CheckDup(Value);
	AddEntry(Value, TYPE_PROC);
	Name = Value;
	Next();
	FormalList();
	SetNumParams(Name, Base);
	k = Decls(SCOPE_LOCAL);
	ProcProlog(Name, k);
	BeginBlock();
	ProcEpilog(Name);
	ClearParams();
}


// Read a single variable
void ReadVar()
{
	CheckIdent();
	CheckTable(Value);
	ReadIt(Value);
	Next();
}


// Process a read statement
void DoRead()
{
	Next();
	MatchString("(");
	ReadVar();
	while (Token == ',')
	{
		Next();
		ReadVar();
	}
	MatchString(")");
}


// Process a write statement
void DoWrite()
{
	int Type;
	
	Next();
	MatchString("(");
	Type = BoolExpression();
	WriteIt(Type);
	while (Token == ',')
	{
		Next();
		Type = BoolExpression();
		WriteIt(Type);
	}
	MatchString(")");
}


// Parse and translate a "begin" block
void BeginBlock()
{
	MatchString(BLOCK_START);
	DoBlock();
	MatchString(BLOCK_END);
}


// Parse and translate a block of statements
void DoBlock()
{
	Scan();
	while ((Token != TOK_END) && (Token != TOK_ELSE))
	{
		switch (Token)
		{
			case TOK_IF:
				DoIf();
				break;
			case TOK_WHILE:
				DoWhile();
				break;
			case TOK_READ:
				DoRead();
				break;
			case TOK_WRITE:
				DoWrite();
				break;
			case TOK_ID:
				AssignOrProc();
				break;
			default:
				Next();
				break;
		}
		MatchSemi();
		Scan();
	}
}


// Allocate storage for a variable
void Alloc (int Type, int Scope)
{
	string Name;
	string SignStr;
	string InitialValue = "0";

	Next();
	if (Token != TOK_ID)
	{
		Expected ("Variable name");
	}
	Name = Value;
	Next();
	if (Token == '=')
	{
		if (Scope == SCOPE_LOCAL)
		{	// FIXME
			Warning ("Sorry, local initializers not implemented");
		}
		Next();
		if (Token == '-')
		{
			SignStr = "-";
			Next();
		}
		if (Token == TOK_NUM)
		{
			InitialValue = Value;
		}
		else
		{
			Abort ("Expected number");
		}
		Next();
	}
	if (Scope == SCOPE_GLOBAL)
	{
		CheckDup(Name);
		AddEntry(Name, Type);
		Allocate(Name, Type, InitialValue);
	}
	else
	{
		AddParam (Name, Type, InitialValue);
	}
}


// 
int Decl(int Type, int Scope)
{
	int DeclCount = 1;
	
	Alloc(Type, Scope);
	while (Token == ',')
	{
		Alloc(Type, Scope);
		DeclCount++;
	}
	return (DeclCount);
}


// Parse and translate declarations
int Decls(int Scope)
{
	int DeclCount = 0;
	
	Scan();
	while ((Token == TOK_BYTE)   ||
		(Token == TOK_UBYTE) ||
		(Token == TOK_WORD)  ||
		(Token == TOK_UWORD) ||
		(Token == TOK_LONG)  ||
		(Token == TOK_ULONG) ||
		(Token == TOK_PROC)  ||
		(Token == TOK_FUNC))
	{
		switch (Token)
		{
			case TOK_BYTE:
				DeclCount += Decl(TYPE_BYTE, Scope);
				break;
			case TOK_UBYTE:
				DeclCount += Decl(TYPE_UBYTE, Scope);
				break;
			case TOK_WORD:
				DeclCount += Decl(TYPE_WORD, Scope);
				break;
			case TOK_UWORD:
				DeclCount += Decl(TYPE_UWORD, Scope);
				break;
			case TOK_LONG:
				DeclCount += Decl(TYPE_LONG, Scope);
				break;
			case TOK_ULONG:
				DeclCount += Decl(TYPE_ULONG, Scope);
				break;
			case TOK_PROC:
				if (Scope == SCOPE_GLOBAL)
				{
					DoProc();
				}
				else
				{
					Abort(string("Nested procdures not allowed."));
				}
				break;
			case TOK_FUNC:
				if (Scope == SCOPE_GLOBAL)
				{
					DoFunc();
				}
				else
				{
					Abort(string("Nested functions not allowed."));
				}
				break;
			
			default:
				Abort(string("TopDecls: Unrecognized keyword: "));
				break;
		}
		MatchSemi();
		Scan();
	}
	return (DeclCount);
}


// Initialize
void Init()
{
	GetChar();
	Next();
}


int main(int argc, char* argv[])
{
	// Try to open an input file if given else read standard input
	if (argc >=2)
	{
		fd = open (argv[1], O_RDONLY);
		if (fd == -1)
		{
			perror("tiny");
			exit(1);
		}
	}
	
	Init();
	Header();
	Decls(SCOPE_GLOBAL);
	Prolog();
	BeginBlock();
	Epilog();
	return (0);
}


